package ostat

import (
	"fmt"
	"math"
)

// from http://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Online_algorithm

const (
	Population = iota
	Sample
)

type OnlineStat struct {
	n    uint64
	mean float64
	m2   float64
	Min  float64
	Max  float64
	typ  uint64
}

func NewSampleStat() *OnlineStat {
	return &OnlineStat{
		Min: math.Inf(1),
		Max: math.Inf(-1),
		typ: Sample,
	}
}

func NewPopulationStat() *OnlineStat {
	return &OnlineStat{
		Min: math.Inf(1),
		Max: math.Inf(-1),
		typ: Population,
	}
}

func MidStreamStat(n uint64, mean, stddev, min, max float64, typ uint64) *OnlineStat {
	return &OnlineStat{
		n:    n,
		mean: mean,
		m2:   stddev * stddev * float64(n),
		Min:  min,
		Max:  max,
		typ:  typ,
	}
}

func (os *OnlineStat) Push(v float64) {
	os.n += 1
	if v < os.Min {
		os.Min = v
	}
	if v > os.Max {
		os.Max = v
	}

	delta := v - os.mean
	os.mean = os.mean + delta/float64(os.n)
	os.m2 = os.m2 + delta*(v-os.mean)
}

func (os *OnlineStat) Mean() float64 {
	if os.n == 0 {
		return 0.0
	}
	return os.mean
}

func (os *OnlineStat) Variance() float64 {
	if os.n == 0 {
		return 0.0
	}
	return os.m2 / float64(os.n-os.typ)
}

func (os *OnlineStat) StdDev() float64 {
	return math.Sqrt(os.Variance())
}

func (os *OnlineStat) String() string {
	return fmt.Sprintf(
		"%+v",
		struct {
			n        uint64
			min      float64
			max      float64
			mean     float64
			variance float64
			stdDev   float64
		}{
			n:        os.n,
			min:      os.Min,
			max:      os.Max,
			mean:     os.Mean(),
			variance: os.Variance(),
			stdDev:   os.StdDev(),
		},
	)
}
